/*
    Copyright (C) 2006  Robert Nelson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "stdafx.h"
#include "defs.h"
#include "ntexe.h"
#include "PEExecutable.h"
#include "cv.h"

#if !defined(O_BINARY)
#define O_BINARY	0
#endif

#if !defined(HAVE_TELL)
inline size_t
tell(int fd)
{
	return lseek(fd, 0, SEEK_CUR);
}
#endif

#if !defined(HAVE_CHSIZE)
#if defined(HAVE_FTRUNCATE)
inline int
chsize(int fd, long size)
{
	return ftruncate(fd, size);
}
#endif
#endif

#define MAX_SYM_LENGTH     255
#define N_FUN              0x24

CPEExecutable::CPEExecutable(void)
   : m_idxDebugSection(-1)
   , m_nSections(0)
   , m_Sections(NULL)
   , m_nStabs(0)
   , m_pStabs(NULL)
   , m_dwStringSize(0)
   , m_pStrings(NULL)
{
}

CPEExecutable::~CPEExecutable(void)
{
   FreeData();
}

void
CPEExecutable::FreeData()
{
   if (m_Sections != NULL)
   {
      delete m_Sections;
   }

   m_nSections = 0;

   if (m_pStabs != NULL)
   {
      delete m_pStabs;
   }

   m_dwStringSize = 0;

   if (m_pStrings != NULL)
   {
      delete m_pStrings;
   }
}

void
CPEExecutable::SetFilename(LPCSTR lpszName)
{
   m_strFileName = lpszName;

   size_t   idxExtension = m_strFileName.find_last_of('.');

   if (idxExtension == std::string::npos)
   {
      m_strFileName.append(".exe");
   }

   size_t   idxBaseName = m_strFileName.find_last_of('\\');

   if (idxBaseName != std::string::npos)
   {
      m_strModuleName = m_strFileName.substr(idxBaseName + 1);
   }
   else
   {
      m_strModuleName = m_strFileName;
   }
}

bool
CPEExecutable::ConvertDebug()
{
   int   fd;

   if ((fd = open(m_strFileName.c_str(), O_RDWR | O_BINARY)) == -1)
   {
      perror(m_strFileName.c_str());
      return false;
   }

   if (ReadNTHeaders(fd) &&
       ReadStabsDebug(fd) &&
       WriteDebugDirectories(fd) &&
       WriteCodeViewData(fd))
   {
      close(fd);
      return true;
   }

   close(fd);

   return false;
}

bool 
CPEExecutable::ReadNTHeaders(int fd)
{
   IMAGE_DOS_HEADER        dosHeader;
   int                     nBytes;

   nBytes = read(fd, &dosHeader, sizeof(dosHeader));

   if (nBytes != sizeof(dosHeader) ||
       dosHeader.e_magic != IMAGE_DOS_SIGNATURE)
   {
      return false;
   }

   m_dwNtHeaderOffset = dosHeader.e_lfanew;

   lseek(fd, m_dwNtHeaderOffset, SEEK_SET);

   nBytes = read(fd, &m_NtHeaders, sizeof(m_NtHeaders));

   if (nBytes != sizeof(m_NtHeaders) ||
       m_NtHeaders.Signature != IMAGE_NT_SIGNATURE)
   {
      return false;
   }

   m_nSections = m_NtHeaders.FileHeader.NumberOfSections;

   if (m_nSections == 0)
   {
      return false;
   }

   m_Sections = new IMAGE_SECTION_HEADER[m_nSections];

   int iSectionsOffset = dosHeader.e_lfanew + 
                         offsetof(IMAGE_NT_HEADERS, OptionalHeader) +
                         m_NtHeaders.FileHeader.SizeOfOptionalHeader;

   lseek(fd, iSectionsOffset, SEEK_SET);

   nBytes = read(fd, m_Sections, m_nSections * sizeof(m_Sections[0]));

   return true;
}

bool 
CPEExecutable::ReadStabsDebug(int fd)
{
   DWORD dwStabOffset = 0, dwStabSize = 0;
   DWORD dwStringOffset = 0, dwStringSize = 0;
   int nBytes;

   m_idxDebugSection = -1;

   if (m_nSections < 3)
   {
      return false;
   }

   if ((strncmp((const char *)m_Sections[m_nSections - 2].Name, ".stab", IMAGE_SIZEOF_SHORT_NAME) != 0 ||
        strncmp((const char *)m_Sections[m_nSections - 1].Name, ".stabstr", IMAGE_SIZEOF_SHORT_NAME) != 0))
   {
      return false;
   }

   dwStabOffset = m_Sections[m_nSections - 2].PointerToRawData;
   dwStabSize = m_Sections[m_nSections - 2].SizeOfRawData;
   dwStringOffset = m_Sections[m_nSections - 1].PointerToRawData;
   dwStringSize = m_Sections[m_nSections - 1].SizeOfRawData;

   if (dwStabSize == 0 || dwStringSize == 0)
   {
      return false;
   }

   m_idxDebugSection = m_nSections - 2;

   STAB_ENTRY  stabHeader;

   lseek(fd, dwStabOffset, SEEK_SET);

   nBytes = read(fd, (void *)&stabHeader, sizeof(stabHeader));

   if (nBytes != sizeof(stabHeader) ||
       stabHeader.byType != 0 ||
       dwStabSize < (stabHeader.wDescription * sizeof(STAB_ENTRY)) ||
       dwStringSize < stabHeader.dwValue)
   {
      return false;
   }

   m_nStabs = stabHeader.wDescription;
   m_pStabs = new STAB_ENTRY[m_nStabs];

   lseek(fd, dwStabOffset, SEEK_SET);

   nBytes = read(fd, (void *)m_pStabs, m_nStabs * sizeof(STAB_ENTRY));

   if ((DWORD)nBytes != (m_nStabs * sizeof(STAB_ENTRY)))
   {
      m_nStabs = 0;
      delete m_pStabs;
      m_pStabs = NULL;

      return false;
   }

   m_dwStringSize = stabHeader.dwValue;
   m_pStrings = new char[m_dwStringSize];

   lseek(fd, dwStringOffset, SEEK_SET);

   nBytes = read(fd, (void *)m_pStrings, m_dwStringSize);

   if ((DWORD)nBytes != m_dwStringSize)
   {
      delete m_pStabs;
      m_pStabs = NULL;
      m_nStabs = 0;

      delete m_pStrings;
      m_pStrings = NULL;
      m_dwStringSize = 0;

      return false;
   }

   return true;
}

bool
CPEExecutable::WriteDebugDirectories(int fd)
{
   m_NtHeaders.FileHeader.NumberOfSymbols = 0;
   m_NtHeaders.FileHeader.PointerToSymbolTable = 0;
   m_NtHeaders.OptionalHeader.CheckSum = 0;

   m_NtHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress =
      m_Sections[m_idxDebugSection].VirtualAddress;
   m_NtHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size =
      sizeof(m_DebugDirs);

   m_NtHeaders.FileHeader.NumberOfSections--;

   DWORD dwNameLength = (DWORD)m_strModuleName.length();
   m_nSections--;

   DWORD dwFileAlignment = m_NtHeaders.OptionalHeader.FileAlignment;
   DWORD dwSectionAlignment = m_NtHeaders.OptionalHeader.SectionAlignment;

   m_Sections[m_idxDebugSection].Characteristics = 
      IMAGE_SCN_CNT_INITIALIZED_DATA | IMAGE_SCN_MEM_DISCARDABLE | IMAGE_SCN_MEM_READ;
   m_Sections[m_idxDebugSection].SizeOfRawData = dwFileAlignment;
   // m_Sections[m_idxDebugSection].VirtualAddress = 0;
   m_Sections[m_idxDebugSection].Misc.VirtualSize = dwSectionAlignment;
   strncpy((char *)m_Sections[m_idxDebugSection].Name, ".debug", IMAGE_SIZEOF_SHORT_NAME);
   m_Sections[m_idxDebugSection].NumberOfLinenumbers = 0;
   m_Sections[m_idxDebugSection].NumberOfRelocations = 0;
   m_Sections[m_idxDebugSection].PointerToLinenumbers = 0;
   // m_Sections[m_idxDebugSection].PointerToRawData = 0;
   m_Sections[m_idxDebugSection].PointerToRelocations = 0;

   memset((void *)&m_DebugDirs[0], 0, sizeof(m_DebugDirs));

   m_DebugDirs[0].PointerToRawData = 
      m_Sections[m_idxDebugSection].PointerToRawData + sizeof(m_DebugDirs);

   m_DebugDirs[0].TimeDateStamp = m_NtHeaders.FileHeader.TimeDateStamp;
   m_DebugDirs[0].Type = IMAGE_DEBUG_TYPE_MISC;

   m_DebugDirs[0].SizeOfData = (sizeof(IMAGE_DEBUG_MISC) + dwNameLength + 3) /4 * 4;

   m_DebugDirs[1].TimeDateStamp = m_NtHeaders.FileHeader.TimeDateStamp;
   m_DebugDirs[1].Type = IMAGE_DEBUG_TYPE_CODEVIEW;

   m_DebugDirs[1].PointerToRawData = 
      m_Sections[m_idxDebugSection].PointerToRawData + m_Sections[m_idxDebugSection].SizeOfRawData;

   m_DebugDirs[1].SizeOfData = 
      sizeof(CVSignature) +
      sizeof(CVDirHeader) + 3 * sizeof(CVDirEntry) +
      offsetof(CVModule, SegInfo) + 
         m_nSections * sizeof(CVSegDesc) + 
         (dwNameLength + 3) / 4 * 4 +
      sizeof(CVSegMap) + m_nSections * sizeof(CVSegMapDesc);

   DWORD  dwDebugMiscLength = (sizeof(IMAGE_DEBUG_MISC) + dwNameLength + 3) / 4 * 4;
   PIMAGE_DEBUG_MISC pDebugMisc = (PIMAGE_DEBUG_MISC)alloca(dwDebugMiscLength);

   memset((void *)pDebugMisc, 0, dwDebugMiscLength);

   pDebugMisc->DataType = IMAGE_DEBUG_MISC_EXENAME;
   pDebugMisc->Length = dwNameLength + 1;
   pDebugMisc->Unicode = FALSE;
   memcpy(&pDebugMisc->Data[0], (void *)m_strModuleName.c_str(), dwNameLength + 1);

   m_NtHeaders.OptionalHeader.SizeOfImage = 
      m_Sections[m_idxDebugSection].VirtualAddress +
      m_Sections[m_idxDebugSection].Misc.VirtualSize;

   lseek(fd, m_dwNtHeaderOffset, SEEK_SET);
   write(fd, &m_NtHeaders, sizeof(m_NtHeaders));

   int iSectionsOffset = m_dwNtHeaderOffset + 
                         offsetof(IMAGE_NT_HEADERS, OptionalHeader) +
                         m_NtHeaders.FileHeader.SizeOfOptionalHeader +
                         (m_idxDebugSection) * sizeof(m_Sections[0]);

   lseek(fd, iSectionsOffset , SEEK_SET);
   write(fd, (void *)&m_Sections[m_idxDebugSection], sizeof(m_Sections[0]));

   lseek(fd, m_Sections[m_idxDebugSection].PointerToRawData, SEEK_SET);
   write(fd, (void *)&m_DebugDirs, sizeof(m_DebugDirs));

   write(fd, (void *)pDebugMisc, dwDebugMiscLength);

   return true;
}

bool
CPEExecutable::WriteCodeViewData(int fd)
{
   CVSignature    Signature = { { 'N', 'B', '0', '9' }, sizeof(Signature) };
   int            iSignatureOffset;
   
   iSignatureOffset = m_Sections[m_idxDebugSection].PointerToRawData + m_Sections[m_idxDebugSection].SizeOfRawData;
   lseek(fd, iSignatureOffset, SEEK_SET);

   write(fd, (void *)&Signature, sizeof(Signature));

   CVDirHeader    DirHeader;
   CVDirEntry     DirEntry;

   memset((void *)&DirHeader, 0, sizeof(DirHeader));
   DirHeader.cbDirHeader = sizeof(DirHeader);
   DirHeader.cbDirEntry = sizeof(DirEntry);
   DirHeader.cDir = 3;
   // DirHeader.lfoNextDir = 0;
   // DirHeader.flags = 0;

   write(fd, (void *)&DirHeader, sizeof(DirHeader));

   DirEntry.SubSection = sstModule;
   DirEntry.iMod = 0xFFFF;
   DirEntry.lfo = sizeof(Signature) +
                  sizeof(DirHeader) + 3 * sizeof(DirEntry);
   DirEntry.cb =  (DWORD)(offsetof(CVModule, SegInfo) + 
                  m_nSections * sizeof(CVSegDesc) +
                  (m_strModuleName.length() + 4) / 4 * 4);

   write(fd, (void *)&DirEntry, sizeof(DirEntry));

   DirEntry.SubSection = sstSegMap;
   DirEntry.lfo += DirEntry.cb;
   DirEntry.cb =  sizeof(CVSegMap) + m_nSections * sizeof(CVSegMapDesc);

   write(fd, (void *)&DirEntry, sizeof(DirEntry));

   DirEntry.SubSection = sstGlobalPub;
   DirEntry.lfo += DirEntry.cb;
   DirEntry.cb =  0;

   int   iGlobalPubDirOffset = tell(fd);

   write(fd, (void *)&DirEntry, sizeof(DirEntry));

   CVModule    Module;

   memset((void *)&Module, 0, sizeof(Module));

   // Module.ovlNumber = 0;
   // Module.iLib = 0;
   Module.cSeg = (USHORT)m_nSections;
   Module.Style[0] = 'C';
   Module.Style[1] = 'V';

   write(fd, (void *)&Module, offsetof(CVModule, SegInfo));

   CVSegDesc   SegDesc;

   memset((void *)&SegDesc, 0, sizeof(SegDesc));

   for (int index = 0; index < m_nSections; index++)
   {
      SegDesc.Seg = (USHORT)(index + 1);
      // SegDesc.pad = 0;
      // SegDesc.Off = 0;
      SegDesc.cbSeg = m_Sections[index].Misc.VirtualSize;

      write(fd, (void *)&SegDesc, sizeof(SegDesc));
   }

   char szName[MAX_SYM_LENGTH];

   DWORD   dwNameLength = (DWORD)m_strModuleName.length();

   szName[0] = (char)dwNameLength;
   memcpy((void *)&szName[1], (void *)m_strModuleName.c_str(), dwNameLength);
   memset((void *)&szName[1 + dwNameLength], 0, (dwNameLength + 4) / 4 * 4 - dwNameLength - 1);

   write(fd, (void *)&szName[0], (dwNameLength + 4) / 4 * 4);

   CVSegMap       SegMap;

   memset((void *)&SegMap, 0, sizeof(SegMap));

   SegMap.cSeg = (USHORT)m_nSections;
   SegMap.cSegLog = (USHORT)m_nSections;

   write(fd, (void *)&SegMap, sizeof(SegMap));

   CVSegMapDesc   SegMapDesc;

   memset(&SegMapDesc, 0, sizeof(SegMapDesc));

   for (int index = 0; index < m_nSections; index++)
   {
      // SegMapDesc.flags = 0;
      // SegMapDesc.ovl = 0;
      // SegMapDesc.group = 0;
      SegMapDesc.frame = (USHORT)(index + 1);
      SegMapDesc.iSegName = 0xFFFF;
      SegMapDesc.iClassName = 0xFFFF;
      // SegMapDesc.offset = 0;
      SegMapDesc.cbSeg = m_Sections[index].Misc.VirtualSize;

      write(fd, (void *)&SegMapDesc, sizeof(SegMapDesc));
   }

   CVSymHash      SymHash;

   memset((void *)&SymHash, 0, sizeof(SymHash));

   // SymHash.symhash = 0;
   // SymHash.addrhash = 0;
   // SymHash.cbSymbol = 0;
   // SymHash.cbHSym = 0;
   // SymHash.cbHAddr = 0;

   int            iSymHashOffset = tell(fd);

   write(fd, (void *)&SymHash, sizeof(SymHash));

   DWORD dwSymbolSize = WriteCodeViewSymbols(fd);

   lseek(fd, iSymHashOffset + offsetof(CVSymHash, cbSymbol), SEEK_SET);
   write(fd, (void *)&dwSymbolSize, sizeof(dwSymbolSize));

   dwSymbolSize += sizeof(SymHash);

   lseek(fd, iGlobalPubDirOffset + offsetof(CVDirEntry, cb), SEEK_SET);
   write(fd, (void *)&dwSymbolSize, sizeof(dwSymbolSize));

   m_DebugDirs[1].SizeOfData += dwSymbolSize;

   lseek(fd, m_Sections[m_idxDebugSection].PointerToRawData + sizeof(m_DebugDirs[0]) + offsetof(IMAGE_DEBUG_DIRECTORY, SizeOfData), SEEK_SET);
   write(fd, (void *)&m_DebugDirs[1].SizeOfData, sizeof(m_DebugDirs[1].SizeOfData));

   lseek(fd, m_DebugDirs[1].PointerToRawData + m_DebugDirs[1].SizeOfData, SEEK_SET);

   Signature.filepos = (long)~(tell(fd) - iSignatureOffset + sizeof(Signature));

   write(fd, (void *)&Signature, sizeof(Signature));

   chsize(fd, tell(fd));

   return true;
}

DWORD
CPEExecutable::WriteCodeViewSymbols(int fd)
{
   long        acPublicData[(sizeof(PUBSYM32) + MAX_SYM_LENGTH + 3) / 4];
   PUBSYM32   *psymPublic = (PUBSYM32 *)&acPublicData[0];
   int         nSectionSize = 0;

   for (DWORD index = 0; index < m_nStabs; index++)
   {
      if (m_pStabs[index].byType == N_FUN)
      {
         if (m_pStabs[index].dwString != 0)
         {
            if (m_pStabs[index].dwString >= m_dwStringSize)
            {
               return 0;
            }

            const char *pszString = m_pStrings + m_pStabs[index].dwString;
            const char *pColon = strchr(pszString, ':');

            if (pColon == NULL)
            {
               return 0;
            }

            short        nStringLength = (short)(pColon - pszString);

            if (nStringLength > (MAX_SYM_LENGTH - 1))
            {
               nStringLength = MAX_SYM_LENGTH - 1;
            }

            USHORT      usPublicSize = (USHORT)((sizeof(PUBSYM32) + nStringLength + 4) / 4 * 4);

            memset(psymPublic, 0, usPublicSize);

            psymPublic->reclen = usPublicSize - 2;
            psymPublic->rectyp = S_PUB32;
            psymPublic->off = 
               m_pStabs[index].dwValue - 
               m_NtHeaders.OptionalHeader.ImageBase - 
               m_Sections[0].VirtualAddress;
            psymPublic->seg = 1;
            psymPublic->typind = 0;
            psymPublic->name[0] = (UCHAR)(nStringLength + 1);
            psymPublic->name[1] = '_';
            memcpy(&psymPublic->name[2], pszString, nStringLength);

            write(fd, psymPublic, usPublicSize);

            nSectionSize += usPublicSize;
         }
      }
   }

   return nSectionSize;
}
